/**
 * FFT Waterfall Display
 * Uses I2S ICS-43434 microphone input to visualize audio frequencies
 * TFT_eSPI using Setup24_ST7789.h
 */

#include <Arduino.h>
#include <TFT_eSPI.h>
#include "arduinoFFT.h"
#include <driver/i2s.h>
#include <Adafruit_NeoPixel.h>

// NeoPixel
//////////////////////////////////////////////////////////////////
#define NEOPIXEL_PIN    46
#define NEOPIXEL_COUNT  1
Adafruit_NeoPixel strip(NEOPIXEL_COUNT, NEOPIXEL_PIN, NEO_GRB + NEO_KHZ800);
float colorPhase = 0.0;
const float COLOR_CYCLE_SPEED = 0.02; // how fast to cycle neopixel
float smoothedMaxValue = 0.0;
const float BRIGHTNESS_SMOOTHING = 0.85;

void heatmapToRGB(float phase, uint8_t &r, uint8_t &g, uint8_t &b) {
  uint8_t value = (uint8_t)(phase * 255.0);
  if (value < 32) {
    float t = value / 31.0;
    r = 13 + 47 * t;
    g = 8 + 9 * t;
    b = 135 + 19 * t;
  } else if (value < 96) {
    float t = (value - 32) / 63.0;
    r = 60 + 140 * t;
    g = 17 + 15 * t;
    b = 154 - 19 * t;
  } else if (value < 160) {
    float t = (value - 96) / 63.0;
    r = 200 + 45 * t;
    g = 32 + 108 * t;
    b = 135 - 73 * t;
  } else {
    float t = (value - 160) / 95.0;
    r = 245 + 10 * t;
    g = 140 + 115 * t;
    b = 62 + 69 * t;
  }
}

void updateNeoPixel(float maxAmplitude) {
  colorPhase += COLOR_CYCLE_SPEED;
  if (colorPhase >= 1.0) {
    colorPhase = 0.0;
  }
  smoothedMaxValue = BRIGHTNESS_SMOOTHING * smoothedMaxValue + (1.0 - BRIGHTNESS_SMOOTHING) * maxAmplitude;
  uint8_t r, g, b;
  heatmapToRGB(colorPhase, r, g, b);
  float brightness = constrain(smoothedMaxValue * autoGainLevel * 0.5, 0.0, 1.0);
  brightness = sqrt(brightness);
  brightness = max(brightness, 0.05f);
  r = (uint8_t)(r * brightness);
  g = (uint8_t)(g * brightness);
  b = (uint8_t)(b * brightness);
  strip.setPixelColor(0, strip.Color(r, g, b));
  strip.show();
}


// Display
//////////////////////////////////////////////////////////////////
#define TFT_WIDTH       320
#define TFT_HEIGHT      170
#define WATERFALL_HEIGHT  140
#define SPECTRUM_HEIGHT   40
#define HEADER_HEIGHT     25
#define HEADER_FONT       &Orbitron_Light_24
#define TFT_TRANSPARENT   0x0000
TFT_eSPI tft = TFT_eSPI();
TFT_eSprite mainSprite = TFT_eSprite(&tft);
TFT_eSprite waterfallSprite = TFT_eSprite(&tft);
TFT_eSprite headerSprite = TFT_eSprite(&tft);

// I2S and Audio
//////////////////////////////////////////////////////////////////
#define I2S_WS          17    // LRCLK
#define I2S_SD          16    // DOUT
#define I2S_SCK         18    // BCLK
#define I2S_PORT        I2S_NUM_0

#define SAMPLES         256     // Must be a power of 2?
#define SAMPLING_FREQ   16000   // Hz (I2S can handle higher frequencies)
#define I2S_SAMPLE_RATE 16000
#define I2S_SAMPLE_BITS 32
#define I2S_READ_LEN    (2 * SAMPLES)

#define MAX_DISPLAY_FREQ  4000
#define MIN_DISPLAY_FREQ  100

#define FREQ_BIN_SIZE     ((float)SAMPLING_FREQ / SAMPLES)  // Hz per bin (~31.25 Hz with 512 samples)
#define MIN_FREQ_BIN      ((int)(MIN_DISPLAY_FREQ / FREQ_BIN_SIZE))
#define MAX_FREQ_BIN      ((int)(MAX_DISPLAY_FREQ / FREQ_BIN_SIZE))
#define DISPLAY_FREQ_BINS (MAX_FREQ_BIN - MIN_FREQ_BIN + 1)

// High-pass filter configuration
#define HPF_CUTOFF_FREQ   80.0
#define HPF_ALPHA         0.95
float hpf_prev_input = 0.0;
float hpf_prev_output = 0.0;
float dc_offset = 0.0;
float dc_offset_alpha = 0.999;

// Sensitivity and scaling controls
#define SENSITIVITY_SCALE   0.1
#define MIN_THRESHOLD      0.00005
#define MAX_SCALE_FACTOR   5.0
#define AUTO_GAIN_FACTOR   0.95

// Low frequency suppression
#define LOW_FREQ_SUPPRESS_FACTOR  0.3

// FFT data arrays
double vReal[SAMPLES];
double vImag[SAMPLES];
ArduinoFFT<double> FFT(vReal, vImag, SAMPLES, SAMPLING_FREQ);
double displayFreqData[DISPLAY_FREQ_BINS];

// Display parameters
const float barWidth = (float)TFT_WIDTH / DISPLAY_FREQ_BINS;
uint8_t frequencyHistory[WATERFALL_HEIGHT][DISPLAY_FREQ_BINS];
int historyIndex = 0;
int32_t i2s_read_buff[I2S_READ_LEN];

// Auto gain control
float autoGainLevel = 1.0;
float maxSignalLevel = 0.0;

float highPassFilter(float input) {
  float output = HPF_ALPHA * (hpf_prev_output + input - hpf_prev_input);
  hpf_prev_input = input;
  hpf_prev_output = output;
  return output;
}

float removeDCOffset(float input) {
  dc_offset = dc_offset_alpha * dc_offset + (1.0 - dc_offset_alpha) * input;
  return input - dc_offset;
}

void i2s_install() {
  const i2s_config_t i2s_config = {
    .mode = i2s_mode_t(I2S_MODE_MASTER | I2S_MODE_RX),
    .sample_rate = I2S_SAMPLE_RATE,
    .bits_per_sample = i2s_bits_per_sample_t(I2S_SAMPLE_BITS),
    .channel_format = I2S_CHANNEL_FMT_ONLY_LEFT,
    .communication_format = i2s_comm_format_t(I2S_COMM_FORMAT_STAND_I2S),
    .intr_alloc_flags = 0,
    .dma_buf_count = 2,
    .dma_buf_len = 128,
    .use_apll = false
  };

  i2s_driver_install(I2S_PORT, &i2s_config, 0, NULL);
}

void i2s_setpin() {
  const i2s_pin_config_t pin_config = {
    .bck_io_num = I2S_SCK,
    .ws_io_num = I2S_WS,
    .data_out_num = I2S_PIN_NO_CHANGE,
    .data_in_num = I2S_SD
  };

  i2s_set_pin(I2S_PORT, &pin_config);
}

uint16_t valueToColor(uint8_t value) {
  uint8_t r, g, b;
  if (value < 32) {
    float t = value / 31.0;
    r = 13 + 47 * t;
    g = 8 + 9 * t;
    b = 135 + 19 * t;
  } else if (value < 96) {
    float t = (value - 32) / 63.0;
    r = 60 + 140 * t;
    g = 17 + 15 * t;
    b = 154 - 19 * t;
  } else if (value < 160) {
    float t = (value - 96) / 63.0;
    r = 200 + 45 * t;
    g = 32 + 108 * t;
    b = 135 - 73 * t;
  } else {
    float t = (value - 160) / 95.0;
    r = 245 + 10 * t;
    g = 140 + 115 * t;
    b = 62 + 69 * t;
  }
  
  return tft.color565(r, g, b);
}

void extractDisplayFrequencies() {
  for (int i = 0; i < DISPLAY_FREQ_BINS; i++) {
    int fftBin = MIN_FREQ_BIN + i;
    if (fftBin < SAMPLES / 2) {
      displayFreqData[i] = vReal[fftBin];
      float freq = (MIN_FREQ_BIN + i) * FREQ_BIN_SIZE;
      if (freq < 200.0) {
        float suppression = LOW_FREQ_SUPPRESS_FACTOR + 
                           (1.0 - LOW_FREQ_SUPPRESS_FACTOR) * (freq / 200.0);
        displayFreqData[i] *= suppression;
      }
    } else {
      displayFreqData[i] = 0;
    }
  }
  float currentMaxValue = 0;
  for (int i = 0; i < DISPLAY_FREQ_BINS; i++) {
    if (displayFreqData[i] > currentMaxValue) {
      currentMaxValue = displayFreqData[i];
    }
  }
  if (currentMaxValue > maxSignalLevel) {
    maxSignalLevel = currentMaxValue;
  } else {
    maxSignalLevel = maxSignalLevel * AUTO_GAIN_FACTOR + currentMaxValue * (1.0 - AUTO_GAIN_FACTOR);
  }
  
  // auto gain
  if (maxSignalLevel > 0.001) {
    autoGainLevel = MAX_SCALE_FACTOR / maxSignalLevel;
  } else {
    autoGainLevel = 1.0;
  }

  autoGainLevel = constrain(autoGainLevel, 0.1, 100.0);
  for (int i = 0; i < DISPLAY_FREQ_BINS; i++) {
    float value = displayFreqData[i];
    if (value < MIN_THRESHOLD) {
      value = 0;
    }
    if (value > 0) {
      value = value * autoGainLevel;
      value = log10(1 + value * 9) / log10(10);
    }
    uint8_t intensity = constrain(value * 255, 0, 255);
    frequencyHistory[historyIndex][i] = intensity;
  }
}

float captureAndProcessAudio() {
  size_t bytesIn = 0;
  // Read I2S data
  esp_err_t result = i2s_read(I2S_PORT, &i2s_read_buff, I2S_READ_LEN * sizeof(int32_t), &bytesIn, portMAX_DELAY);
  if (result == ESP_OK) {
    int16_t samples_read = bytesIn / sizeof(int32_t);
    for (int i = 0; i < SAMPLES && i < samples_read; i++) {
      int32_t sample = i2s_read_buff[i] >> 8;
      float audioSample = ((float)sample / 8388608.0) * SENSITIVITY_SCALE;
      audioSample = removeDCOffset(audioSample);
      audioSample = highPassFilter(audioSample);
      vReal[i] = audioSample;
      vImag[i] = 0;
    }
    for (int i = samples_read; i < SAMPLES; i++) {
      vReal[i] = 0;
      vImag[i] = 0;
    }
  } else {
    for (int i = 0; i < SAMPLES; i++) {
      vReal[i] = 0;
      vImag[i] = 0;
    }
  }
  FFT.windowing(FFT_WIN_TYP_HAMMING, FFT_FORWARD);
  FFT.compute(FFT_FORWARD);
  FFT.complexToMagnitude();
  extractDisplayFrequencies();
  float currentMaxValue = 0;
  for (int i = 0; i < DISPLAY_FREQ_BINS; i++) {
    if (displayFreqData[i] > currentMaxValue) {
      currentMaxValue = displayFreqData[i];
    }
  }
  
  return currentMaxValue;
}

void drawFrequencySpectrum() {
  for (int i = 0; i < DISPLAY_FREQ_BINS; i++) {
    float value = displayFreqData[i];
    
    if (value < MIN_THRESHOLD) value = 0;
    if (value > 0) {
      value = value * autoGainLevel;
      value = log10(1 + value * 9) / log10(10);
    }
    
    int barHeight = value * (SPECTRUM_HEIGHT + HEADER_HEIGHT);
    barHeight = constrain(barHeight, 0, SPECTRUM_HEIGHT + HEADER_HEIGHT);

    float x = i * barWidth;
    uint16_t color;
    if (value > 0.7) {
      color = TFT_RED;
    } else if (value > 0.4) {
      color = TFT_ORANGE;
    } else if (value > 0.2) {
      color = TFT_YELLOW;
    } else {
      color = TFT_DARKGREY;
    }
    if (barHeight > 0) {
      mainSprite.fillRect(x, HEADER_HEIGHT + SPECTRUM_HEIGHT - barHeight, 
                         barWidth - 1, barHeight, color);
    }
  }
}

void drawWaterfall() {
  waterfallSprite.fillSprite(TFT_BLACK);
  
  for (int y = 0; y < WATERFALL_HEIGHT; y++) {
    int histLine = (historyIndex + WATERFALL_HEIGHT - y) % WATERFALL_HEIGHT;
    
    for (int x = 0; x < DISPLAY_FREQ_BINS; x++) {
      uint8_t value = frequencyHistory[histLine][x];
      // if (value <= 0) {value=0;}
      if (value > 0) {
        float xPos = x * barWidth;
        waterfallSprite.fillRect(xPos, y, barWidth, 1, valueToColor(value));
      }
    }
  }
  waterfallSprite.pushToSprite(&mainSprite, 0, HEADER_HEIGHT + SPECTRUM_HEIGHT, TFT_BLACK);
}

void drawHeader(float maxValue) {
  headerSprite.fillSprite(TFT_BLACK);
  headerSprite.setFreeFont(HEADER_FONT);  
  headerSprite.setCursor(10, 25);
  headerSprite.println("FFT Waterfall");
  headerSprite.setFreeFont();
  headerSprite.setTextSize(1);
  headerSprite.setCursor(213, 18);
  headerSprite.print("Max:");
  headerSprite.print(maxValue, 3);
  headerSprite.pushToSprite(&mainSprite, 0, 0, TFT_BLACK);
}

void setup() {
  // Serial.begin(115200);

  // NeoPixel
  strip.begin();
  strip.show();
  strip.setBrightness(64);

  // I2S microphone
  i2s_install();
  i2s_setpin();
  i2s_start(I2S_PORT);
  
  delay(500);

  // Display
  pinMode(TFT_BL, OUTPUT);
  digitalWrite(TFT_BL, HIGH);

  tft.init();
  tft.setRotation(1);
  tft.fillScreen(TFT_BLACK);
  
  mainSprite.setColorDepth(8);
  mainSprite.createSprite(TFT_WIDTH, TFT_HEIGHT);
  mainSprite.setTextColor(TFT_WHITE, TFT_BLACK);
  waterfallSprite.setColorDepth(8);
  waterfallSprite.createSprite(TFT_WIDTH, WATERFALL_HEIGHT);
  waterfallSprite.fillSprite(TFT_BLACK);

  headerSprite.setColorDepth(8);
  headerSprite.createSprite(TFT_WIDTH, HEADER_HEIGHT);
  headerSprite.setTextColor(TFT_WHITE, TFT_BLACK);

  memset(frequencyHistory, 0, sizeof(frequencyHistory));

  // Serial.println("I2S FFT Waterfall Display with NeoPixel initialized");
  // Serial.println("NeoPixel on GPIO: " + String(NEOPIXEL_PIN));
  // Serial.println("Sensitivity Scale: " + String(SENSITIVITY_SCALE));
  // Serial.println("High-Pass Filter Cutoff: " + String(HPF_CUTOFF_FREQ) + " Hz");
  // Serial.println("Low Frequency Suppression Factor: " + String(LOW_FREQ_SUPPRESS_FACTOR));
  // Serial.println("Frequency Range: " + String(MIN_DISPLAY_FREQ) + " - " + String(MAX_DISPLAY_FREQ) + " Hz");
  // Serial.println("Frequency Resolution: " + String(FREQ_BIN_SIZE, 2) + " Hz per bin");
  // Serial.println("Display Bins: " + String(DISPLAY_FREQ_BINS));
  // Serial.println("Bar Width: " + String(barWidth, 2) + " pixels");
  
  // Serial.println("\nMemory Usage:");
  // Serial.println("FFT Arrays (vReal + vImag): " + String(2 * SAMPLES * sizeof(double)) + " bytes");
  // Serial.println("Display Freq Data: " + String(DISPLAY_FREQ_BINS * sizeof(double)) + " bytes");
  // Serial.println("Frequency History: " + String(WATERFALL_HEIGHT * DISPLAY_FREQ_BINS) + " bytes");
  // Serial.println("I2S Buffer: " + String(I2S_READ_LEN * sizeof(int32_t)) + " bytes");
}

void loop() {
  float maxValue = captureAndProcessAudio();
  updateNeoPixel(maxValue);
  mainSprite.fillSprite(TFT_BLACK);
  drawFrequencySpectrum();
  drawWaterfall();
  drawHeader(maxValue);
  mainSprite.pushSprite(0, 0);
  historyIndex = (historyIndex + 1) % WATERFALL_HEIGHT;
  delay(20);
}